import { SourceTable } from '@powersync/service-core';
import { PgChunk, PgConnection, PgTypeOid } from '@powersync/service-jpgwire';
import { SqliteValue } from '@powersync/service-sync-rules';
export interface SnapshotQuery {
    initialize(): Promise<void>;
    /**
     * Returns an async iterable iterator that yields chunks of data.
     *
     * If the last chunk has 0 rows, it indicates that there are no more rows to fetch.
     */
    nextChunk(): AsyncIterableIterator<PgChunk>;
}
export type PrimaryKeyValue = Record<string, SqliteValue>;
export interface MissingRow {
    table: SourceTable;
    key: PrimaryKeyValue;
}
/**
 * Snapshot query using a plain SELECT * FROM table; chunked using
 * DELCLARE CURSOR / FETCH.
 *
 * This supports all tables, but does not efficiently resume the snapshot
 * if the process is restarted.
 */
export declare class SimpleSnapshotQuery implements SnapshotQuery {
    private readonly connection;
    private readonly table;
    private readonly chunkSize;
    constructor(connection: PgConnection, table: SourceTable, chunkSize?: number);
    initialize(): Promise<void>;
    nextChunk(): AsyncIterableIterator<PgChunk>;
}
/**
 * Performs a table snapshot query, chunking by ranges of primary key data.
 *
 * This may miss some rows if they are modified during the snapshot query.
 * In that case, logical replication will pick up those rows afterwards,
 * possibly resulting in an IdSnapshotQuery.
 *
 * Currently, this only supports a table with a single primary key column,
 * of a select few types.
 */
export declare class ChunkedSnapshotQuery implements SnapshotQuery {
    private readonly connection;
    private readonly table;
    private readonly chunkSize;
    /**
     * Primary key types that we support for chunked snapshots.
     *
     * Can expand this over time as we add more tests,
     * and ensure there are no issues with type conversion.
     */
    static SUPPORTED_TYPES: PgTypeOid[];
    static supports(table: SourceTable): boolean;
    private readonly key;
    lastKey: string | bigint | null;
    constructor(connection: PgConnection, table: SourceTable, chunkSize: number | undefined, lastKeySerialized: Uint8Array | null);
    initialize(): Promise<void>;
    private deserializeKey;
    getLastKeySerialized(): Uint8Array;
    nextChunk(): AsyncIterableIterator<PgChunk>;
}
/**
 * This performs a snapshot query using a list of primary keys.
 *
 * This is not used for general snapshots, but is used when we need to re-fetch specific rows
 * during streaming replication.
 */
export declare class IdSnapshotQuery implements SnapshotQuery {
    private readonly connection;
    private readonly table;
    private readonly keys;
    private didChunk;
    static supports(table: SourceTable): boolean;
    constructor(connection: PgConnection, table: SourceTable, keys: PrimaryKeyValue[]);
    initialize(): Promise<void>;
    nextChunk(): AsyncIterableIterator<PgChunk>;
}
