import { bson } from '@powersync/service-core';
import { PgType, PgTypeOid } from '@powersync/service-jpgwire';
import { escapeIdentifier } from '@powersync/lib-service-postgres';
import { ServiceAssertionError } from '@powersync/lib-services-framework';
/**
 * Snapshot query using a plain SELECT * FROM table; chunked using
 * DELCLARE CURSOR / FETCH.
 *
 * This supports all tables, but does not efficiently resume the snapshot
 * if the process is restarted.
 */
export class SimpleSnapshotQuery {
    connection;
    table;
    chunkSize;
    constructor(connection, table, chunkSize = 10_000) {
        this.connection = connection;
        this.table = table;
        this.chunkSize = chunkSize;
    }
    async initialize() {
        await this.connection.query(`DECLARE snapshot_cursor CURSOR FOR SELECT * FROM ${this.table.escapedIdentifier}`);
    }
    nextChunk() {
        return this.connection.stream(`FETCH ${this.chunkSize} FROM snapshot_cursor`);
    }
}
/**
 * Performs a table snapshot query, chunking by ranges of primary key data.
 *
 * This may miss some rows if they are modified during the snapshot query.
 * In that case, logical replication will pick up those rows afterwards,
 * possibly resulting in an IdSnapshotQuery.
 *
 * Currently, this only supports a table with a single primary key column,
 * of a select few types.
 */
export class ChunkedSnapshotQuery {
    connection;
    table;
    chunkSize;
    /**
     * Primary key types that we support for chunked snapshots.
     *
     * Can expand this over time as we add more tests,
     * and ensure there are no issues with type conversion.
     */
    static SUPPORTED_TYPES = [
        PgTypeOid.TEXT,
        PgTypeOid.VARCHAR,
        PgTypeOid.UUID,
        PgTypeOid.INT2,
        PgTypeOid.INT4,
        PgTypeOid.INT8
    ];
    static supports(table) {
        if (table.replicaIdColumns.length != 1) {
            return false;
        }
        const primaryKey = table.replicaIdColumns[0];
        return primaryKey.typeId != null && ChunkedSnapshotQuery.SUPPORTED_TYPES.includes(Number(primaryKey.typeId));
    }
    key;
    lastKey = null;
    constructor(connection, table, chunkSize = 10_000, lastKeySerialized) {
        this.connection = connection;
        this.table = table;
        this.chunkSize = chunkSize;
        this.key = table.replicaIdColumns[0];
        if (lastKeySerialized != null) {
            this.lastKey = this.deserializeKey(lastKeySerialized);
        }
    }
    async initialize() {
        // No-op
    }
    deserializeKey(key) {
        const decoded = bson.deserialize(key, { useBigInt64: true });
        const keys = Object.keys(decoded);
        if (keys.length != 1) {
            throw new ServiceAssertionError(`Multiple keys found: ${keys.join(', ')}`);
        }
        if (keys[0] != this.key.name) {
            throw new ServiceAssertionError(`Key name mismatch: expected ${this.key.name}, got ${keys[0]}`);
        }
        const value = decoded[this.key.name];
        return value;
    }
    getLastKeySerialized() {
        return bson.serialize({ [this.key.name]: this.lastKey });
    }
    async *nextChunk() {
        let stream;
        const escapedKeyName = escapeIdentifier(this.key.name);
        if (this.lastKey == null) {
            stream = this.connection.stream(`SELECT * FROM ${this.table.escapedIdentifier} ORDER BY ${escapedKeyName} LIMIT ${this.chunkSize}`);
        }
        else {
            if (this.key.typeId == null) {
                throw new Error(`typeId required for primary key ${this.key.name}`);
            }
            let type = Number(this.key.typeId);
            stream = this.connection.stream({
                statement: `SELECT * FROM ${this.table.escapedIdentifier} WHERE ${escapedKeyName} > $1 ORDER BY ${escapedKeyName} LIMIT ${this.chunkSize}`,
                params: [{ value: this.lastKey, type }]
            });
        }
        let primaryKeyIndex = -1;
        for await (let chunk of stream) {
            if (chunk.tag == 'RowDescription') {
                // We get a RowDescription for each FETCH call, but they should
                // all be the same.
                let i = 0;
                const pk = chunk.payload.findIndex((c) => c.name == this.key.name);
                if (pk < 0) {
                    throw new Error(`Cannot find primary key column ${this.key} in results. Keys: ${chunk.payload.map((c) => c.name).join(', ')}`);
                }
                primaryKeyIndex = pk;
            }
            if (chunk.rows.length > 0) {
                this.lastKey = chunk.rows[chunk.rows.length - 1][primaryKeyIndex];
            }
            yield chunk;
        }
    }
}
/**
 * This performs a snapshot query using a list of primary keys.
 *
 * This is not used for general snapshots, but is used when we need to re-fetch specific rows
 * during streaming replication.
 */
export class IdSnapshotQuery {
    connection;
    table;
    keys;
    didChunk = false;
    static supports(table) {
        // We have the same requirements as ChunkedSnapshotQuery.
        // This is typically only used as a fallback when ChunkedSnapshotQuery
        // skipped some rows.
        return ChunkedSnapshotQuery.supports(table);
    }
    constructor(connection, table, keys) {
        this.connection = connection;
        this.table = table;
        this.keys = keys;
    }
    async initialize() {
        // No-op
    }
    async *nextChunk() {
        // Only produce one chunk
        if (this.didChunk) {
            return;
        }
        this.didChunk = true;
        const keyDefinition = this.table.replicaIdColumns[0];
        const ids = this.keys.map((record) => record[keyDefinition.name]);
        const type = PgType.getArrayType(keyDefinition.typeId);
        if (type == null) {
            throw new Error(`Cannot determine primary key array type for ${JSON.stringify(keyDefinition)}`);
        }
        yield* this.connection.stream({
            statement: `SELECT * FROM ${this.table.escapedIdentifier} WHERE ${escapeIdentifier(keyDefinition.name)} = ANY($1)`,
            params: [
                {
                    type: type,
                    value: ids
                }
            ]
        });
    }
}
//# sourceMappingURL=SnapshotQuery.js.map