import { Logger } from '@powersync/lib-services-framework';
import { MetricsEngine, SourceEntityDescriptor, SourceTable, storage } from '@powersync/service-core';
import * as pgwire from '@powersync/service-jpgwire';
import { DatabaseInputRow, SqliteRow, SqlSyncRules, TablePattern } from '@powersync/service-sync-rules';
import { PgManager } from './PgManager.js';
export interface WalStreamOptions {
    logger?: Logger;
    connections: PgManager;
    storage: storage.SyncRulesBucketStorage;
    metrics: MetricsEngine;
    abort_signal: AbortSignal;
    /**
     * Override snapshot chunk length (number of rows), for testing.
     *
     * Defaults to 10_000.
     *
     * Note that queries are streamed, so we don't actually keep that much data in memory.
     */
    snapshotChunkLength?: number;
}
interface InitResult {
    /** True if initial snapshot is not yet done. */
    needsInitialSync: boolean;
    /** True if snapshot must be started from scratch with a new slot. */
    needsNewSlot: boolean;
}
export declare const ZERO_LSN = "00000000/00000000";
export declare const PUBLICATION_NAME = "powersync";
export declare const POSTGRES_DEFAULT_SCHEMA = "public";
export declare const KEEPALIVE_CONTENT = "ping";
export declare const KEEPALIVE_BUFFER: Buffer<ArrayBuffer>;
export declare const KEEPALIVE_STATEMENT: pgwire.Statement;
export declare const isKeepAliveMessage: (msg: pgwire.PgoutputMessage) => boolean;
export declare const sendKeepAlive: (db: pgwire.PgClient) => Promise<void>;
export declare class MissingReplicationSlotError extends Error {
    constructor(message: string);
}
export declare class WalStream {
    sync_rules: SqlSyncRules;
    group_id: number;
    connection_id: number;
    private logger;
    private readonly storage;
    private readonly metrics;
    private readonly slot_name;
    private connections;
    private abort_signal;
    private relationCache;
    private startedStreaming;
    private snapshotChunkLength;
    /**
     * Time of the oldest uncommitted change, according to the source db.
     * This is used to determine the replication lag.
     */
    private oldestUncommittedChange;
    /**
     * Keep track of whether we have done a commit or keepalive yet.
     * We can only compute replication lag if isStartingReplication == false, or oldestUncommittedChange is present.
     */
    private isStartingReplication;
    constructor(options: WalStreamOptions);
    get stopped(): boolean;
    getQualifiedTableNames(batch: storage.BucketStorageBatch, db: pgwire.PgConnection, tablePattern: TablePattern): Promise<storage.SourceTable[]>;
    initSlot(): Promise<InitResult>;
    /**
     * If a replication slot exists, check that it is healthy.
     */
    private checkReplicationSlot;
    estimatedCountNumber(db: pgwire.PgConnection, table: storage.SourceTable): Promise<number>;
    /**
     * Start initial replication.
     *
     * If (partial) replication was done before on this slot, this clears the state
     * and starts again from scratch.
     */
    startInitialReplication(replicationConnection: pgwire.PgConnection, status: InitResult): Promise<void>;
    initialReplication(db: pgwire.PgConnection): Promise<void>;
    static getQueryData(results: Iterable<DatabaseInputRow>): Generator<SqliteRow>;
    private snapshotTableInTx;
    private snapshotTable;
    handleRelation(batch: storage.BucketStorageBatch, descriptor: SourceEntityDescriptor, snapshot: boolean): Promise<SourceTable>;
    /**
     * Process rows that have missing TOAST values.
     *
     * This can happen during edge cases in the chunked intial snapshot process.
     *
     * We handle this similar to an inline table snapshot, but limited to the specific
     * set of rows.
     */
    private resnapshot;
    private getTable;
    writeChange(batch: storage.BucketStorageBatch, msg: pgwire.PgoutputMessage): Promise<storage.FlushedResult | null>;
    replicate(): Promise<void>;
    initReplication(replicationConnection: pgwire.PgConnection): Promise<void>;
    streamChanges(replicationConnection: pgwire.PgConnection): Promise<void>;
    ack(lsn: string, replicationStream: pgwire.ReplicationStream): Promise<void>;
    /**
     * Ensures that the storage is compatible with the replication connection.
     * @throws {DatabaseConnectionError} If the storage is not compatible with the replication connection.
     */
    protected ensureStorageCompatibility(): Promise<storage.ResolvedBucketBatchCommitOptions>;
    /**
     * Check if the replication connection Postgres server supports
     * viewing the contents of logical replication messages.
     */
    protected checkLogicalMessageSupport(): Promise<boolean>;
    getReplicationLagMillis(): Promise<number | undefined>;
}
export {};
